"""Parsers."""

# ./.homeassistant/deps/lib/python/site-packages/rflink/parser.py
# /Library/Frameworks/Python.framework/Versions/3.6//lib/python3.6/site-packages/rflink/parser.py

import logging
import re
import time
from enum import Enum
from typing import Any, Callable, DefaultDict, Dict, Generator, cast

log = logging.getLogger(__name__)

UNKNOWN = "unknown"
SWITCH_COMMAND_TEMPLATE = "{node};{protocol};{id};{switch};{command};"
PACKET_ID_SEP = "_"

# contruct regex to validate packets before parsing
DELIM = ";"
SEQUENCE = "[0-9a-zA-Z]{2}"
PROTOCOL = "[^;]{3,}"
ADDRESS = "[0-9a-zA-Z]+"
BUTTON = "[0-9a-zA-Z]+"
VALUE = "[0-9a-zA-Z]+"
COMMAND = "[0-9a-zA-Z]+"
CONTROL_COMMAND = "[A-Z]+(=[A-Z0-9]+)?"
DATA = "[a-zA-Z0-9;=_]+"
DEBUG_DATA_RTS = "[a-zA-Z0-9;=_ ]+"
DEBUG_DATA = "[a-zA-Z0-9,;=_()]+"
RESPONSES = "OK"
VERSION = r"[0-9a-zA-Z \.-]+"
DEBUG = "DEBUG"
MESSAGE = r"[0-9a-zA-Z \._-]+"

# 10;NewKaku;0cac142;3;ON;
PACKET_COMMAND = DELIM.join(["10", PROTOCOL, ADDRESS, BUTTON, COMMAND])
# 10;MiLightv1;F746;00;3c00;ON;
PACKET_COMMAND2 = DELIM.join(["10", PROTOCOL, ADDRESS, BUTTON, VALUE, COMMAND])
# 10;MERTIK;64;UP;
PACKET_COMMAND3 = DELIM.join(["10", PROTOCOL, ADDRESS, COMMAND])
# 10;DELTRONIC;001c33;
PACKET_COMMAND4 = DELIM.join(["10", PROTOCOL, ADDRESS])

# 10;REBOOT;/10;RTSRECCLEAN=9;
PACKET_CONTROL = DELIM.join(["10", CONTROL_COMMAND])

# 20;D3;OK;
PACKET_RESPONSE = DELIM.join(["20", SEQUENCE, RESPONSES])
# 20;06;NewKaku;ID=008440e6;SWITCH=a;CMD=OFF;
PACKET_DEVICE = DELIM.join(["20", SEQUENCE, PROTOCOL, DATA])
# 20;00;Nodo RadioFrequencyLink - RFLink Gateway V1.1 - R46;
PACKET_VERSION = DELIM.join(["20", SEQUENCE, VERSION])
# 20;00;Internal Pullup on RF-in disabled
PACKET_INFO = DELIM.join(["20", SEQUENCE, MESSAGE])

# 20;75;DEBUG;Pulses=90;Pulses(uSec)=1200,2760,120...
PACKET_DEBUG = DELIM.join(["20", SEQUENCE, DEBUG, DEBUG_DATA])

# 20;01;RFUDEBUG=OFF;
PACKET_RFDEBUGN = DELIM.join(["20", SEQUENCE, "RFDEBUG=ON"])
PACKET_RFDEBUGF = DELIM.join(["20", SEQUENCE, "RFDEBUG=OFF"])
PACKET_RFUDEBUGN = DELIM.join(["20", SEQUENCE, "RFUDEBUG=ON"])
PACKET_RFUDEBUGF = DELIM.join(["20", SEQUENCE, "RFUDEBUG=OFF"])
PACKET_QRFDEBUGN = DELIM.join(["20", SEQUENCE, "QRFDEBUG=ON"])
PACKET_QRFDEBUGF = DELIM.join(["20", SEQUENCE, "QRFDEBUG=OFF"])

PACKET_GPIOON = DELIM.join(["20", SEQUENCE, "setGPIO=ON"])
PACKET_GPIOOFF = DELIM.join(["20", SEQUENCE, "setGPIO=OFF"])

# 20;84;Debug;RTS P1;a63f33003cf000665a5a;
PACKET_DEBUGRTS = DELIM.join(["20", SEQUENCE, "Debug", DEBUG_DATA_RTS])

# 11;20;0B;NewKaku;ID=000005;SWITCH=2;CMD=ON;
PACKET_DEVICE_CREATE = "11;" + PACKET_DEVICE

PACKET_HEADER_RE = (
    "^("
    + "|".join(
        [
            PACKET_VERSION,
            PACKET_DEVICE_CREATE,
            PACKET_RESPONSE,
            PACKET_DEVICE,
            PACKET_COMMAND,
            PACKET_COMMAND2,
            PACKET_COMMAND3,
            PACKET_COMMAND4,
            PACKET_CONTROL,
            PACKET_DEBUG,
            PACKET_INFO,
            PACKET_RFDEBUGN,
            PACKET_RFUDEBUGN,
            PACKET_RFDEBUGF,
            PACKET_RFUDEBUGF,
            PACKET_QRFDEBUGN,
            PACKET_QRFDEBUGF,
            PACKET_GPIOOFF,
            PACKET_GPIOON,
            PACKET_DEBUGRTS,
        ]
    )
    + ");$"
)
packet_header_re = re.compile(PACKET_HEADER_RE)

PacketType = Dict[str, Any]


class PacketHeader(Enum):
    """Packet source identification."""

    master = "10"
    echo = "11"
    gateway = "20"


PACKET_FIELDS = {
    "awinsp": "average_windspeed",
    "baro": "barometric_pressure",
    "bat": "battery",
    "bforecast": "weather_forecast",
    "chime": "doorbell_melody",
    "cmd": "command",
    "co2": "co2_air_quality",
    "current": "current_phase_1",
    "current2": "current_phase_2",
    "current3": "current_phase_3",
    "dist": "distance",
    "fw": "firmware",
    "hstatus": "humidity_status",
    "hum": "humidity",
    "hw": "hardware",
    "kwatt": "kilowatt",
    "lux": "light_intensity",
    "meter": "meter_value",
    "rain": "total_rain",
    "rainrate": "rain_rate",
    "raintot": "total_rain",
    "rev": "revision",
    "sound": "noise_level",
    "temp": "temperature",
    "uv": "uv_intensity",
    "ver": "version",
    "volt": "voltage",
    "watt": "watt",
    "winchl": "windchill",
    "windir": "winddirection",
    "wings": "windgusts",
    "winsp": "windspeed",
    "wintmp": "windtemp",
}

UNITS = {
    "awinsp": "km/h",
    # depends on sensor
    "baro": None,
    "bat": None,
    "bforecast": None,
    "chime": None,
    "cmd": None,
    "co2": None,
    "current": "A",
    "current2": "A",
    "current3": "A",
    # depends on sensor
    "dist": None,
    "fw": None,
    "hstatus": None,
    "hum": "%",
    "hw": None,
    "kwatt": "kW",
    "lux": "lux",
    # depends on sensor
    "meter": None,
    "rain": "mm",
    "rainrate": "mm",
    "raintot": "mm",
    "rev": None,
    # unknown, might be dB?
    "sound": None,
    # might be °F, but default to something
    "temp": "°C",
    "uv": None,
    "ver": None,
    "volt": "v",
    "watt": "w",
    "winchl": "°C",
    "windir": "°",
    "wings": "km/h",
    "winsp": "km/h",
    "wintmp": "°C",
}

HSTATUS_LOOKUP = {
    "0": "normal",
    "1": "comfortable",
    "2": "dry",
    "3": "wet",
}
BFORECAST_LOOKUP = {
    "0": "no_info",
    "1": "sunny",
    "2": "partly_cloudy",
    "3": "cloudy",
    "4": "rain",
}


def signed_to_float(hex: str) -> float:
    """Convert signed hexadecimal to floating value."""
    if int(hex, 16) & 0x8000:
        return -(int(hex, 16) & 0x7FFF) / 10
    else:
        return int(hex, 16) / 10


VALUE_TRANSLATION = cast(
    Dict[str, Callable[[str], str]],
    {
        "awinsp": lambda hex: int(hex, 16) / 10,
        "baro": lambda hex: int(hex, 16),
        "bforecast": lambda x: BFORECAST_LOOKUP.get(x, "Unknown"),
        "chime": int,
        "co2": int,
        "current": int,
        "current2": int,
        "current3": int,
        "dist": int,
        "hstatus": lambda x: HSTATUS_LOOKUP.get(x, "Unknown"),
        "hum": int,
        "kwatt": lambda hex: int(hex, 16),
        "lux": lambda hex: int(hex, 16),
        "meter": int,
        "rain": lambda hex: int(hex, 16) / 10,
        "rainrate": lambda hex: int(hex, 16) / 10,
        "raintot": lambda hex: int(hex, 16) / 10,
        "sound": int,
        "temp": signed_to_float,
        "uv": lambda hex: int(hex, 16),
        "volt": int,
        "watt": lambda hex: int(hex, 16),
        "winchl": signed_to_float,
        "windir": lambda windir: int(windir) * 22.5,
        "wings": lambda hex: int(hex, 16) / 10,
        "winsp": lambda hex: int(hex, 16) / 10,
        "wintmp": signed_to_float,
    },
)


BANNER_RE = (
    r"(?P<hardware>[a-zA-Z\s]+) - (?P<firmware>[a-zA-Z\s]+) "
    r"V(?P<version>[0-9\.]+) - R(?P<revision>[0-9\.]+)"
)


def valid_packet(packet: str) -> bool:
    """Verify if packet is valid.

    >>> valid_packet('20;08;UPM/Esic;ID=1003;RAIN=0010;BAT=OK;')
    True
    >>> # invalid packet due to leftovers in serial buffer
    >>> valid_packet('20;00;N20;00;Nodo RadioFrequencyLink - RFLink Gateway V1.1 - R45')
    False
    """
    return bool(packet_header_re.match(packet))


def decode_packet(packet: str) -> PacketType:
    """Break packet down into primitives, and do basic interpretation.

    >>> decode_packet('20;06;Kaku;ID=41;SWITCH=1;CMD=ON;') == {
    ...     'node': 'gateway',
    ...     'protocol': 'kaku',
    ...     'id': '000041',
    ...     'switch': '1',
    ...     'command': 'on',
    ... }
    True
    """
    node_id, _, protocol, attrs = packet.split(DELIM, 3)

    data = cast(PacketType, {"node": PacketHeader(node_id).name})

    # make exception for version response
    data["protocol"] = UNKNOWN
    if "=" in protocol:
        attrs = protocol + DELIM + attrs

    # no attributes but instead the welcome banner
    elif "RFLink Gateway" in protocol:
        data.update(parse_banner(protocol))

    elif protocol == "PONG":
        data["ping"] = protocol.lower()

    # debug response
    elif protocol.lower() == "debug":
        data["protocol"] = protocol.lower()
        if attrs.startswith("RTS P1"):
            data["rts_p1"] = attrs.strip(DELIM).split(DELIM)[1]
        else:
            data["tm"] = packet[3:5]

    # failure response
    elif protocol == "CMD UNKNOWN":
        data["response"] = "command_unknown"
        data["ok"] = False

    # ok response
    elif protocol == "OK":
        data["ok"] = True

    # generic message from gateway
    elif node_id == "20" and not attrs:
        data["message"] = protocol

    # its a regular packet
    else:
        data["protocol"] = protocol.lower()

    # convert key=value pairs where needed
    for attr in filter(None, attrs.strip(DELIM).split(DELIM)):
        if "=" not in attr:
            continue
        key, value = attr.lower().split("=", 1)
        if key in VALUE_TRANSLATION:
            try:
                value = VALUE_TRANSLATION[key](value)
            except ValueError:
                log.warning(
                    "Could not convert attr '%s' value '%s' to expected type '%s'",
                    key,
                    value,
                    VALUE_TRANSLATION[key].__name__,
                )
                continue
        name = PACKET_FIELDS.get(key, key)
        data[name] = value
        unit = UNITS.get(key, None)

        if unit:
            data[name + "_unit"] = unit

    # correct KaKu device address
    if data.get("protocol", "") == "kaku" and len(data["id"]) != 6:
        data["id"] = "0000" + data["id"]

    return data


def parse_banner(banner: str) -> Dict[str, str]:
    """Extract hardware/firmware name and version from banner."""
    match = re.match(BANNER_RE, banner)
    return match.groupdict() if match else {}


def encode_packet(packet: PacketType) -> str:
    """Construct packet string from packet dictionary.

    >>> encode_packet({
    ...     'protocol': 'newkaku',
    ...     'id': '000001',
    ...     'switch': '01',
    ...     'command': 'on',
    ... })
    '10;newkaku;000001;01;on;'
    """
    if packet["protocol"] == "rfdebug":
        return "10;RFDEBUG=%s;" % packet["command"]
    elif packet["protocol"] == "rfudebug":
        return "10;RFUDEBUG=%s;" % packet["command"]
    elif packet["protocol"] == "qrfdebug":
        return "10;QRFDEBUG=%s;" % packet["command"]
    else:
        return SWITCH_COMMAND_TEMPLATE.format(node=PacketHeader.master.value, **packet)


# create lookup table of not easy to reverse protocol names
translate_protocols = [
    "Ikea Koppla",
    "Alecto V1",
    "Alecto V2",
    "UPM/Esic",
    "Oregon TempHygro",
    "Oregon TempHygro",
    "Oregon BTHR",
    "Oregon Rain",
    "Oregon Rain2",
    "Oregon Wind",
    "Oregon Wind2",
    "Oregon UVN128/138",
    "Plieger York",
    "Byron SX",
    "CAME-TOP432",
]


class TranslationsDict(DefaultDict[str, str]):
    """Generate translations for Rflink protocols to serializable names."""

    def __missing__(self, key: str) -> str:
        """If translation does not exist yet add it and its reverse."""
        value = re.sub(r"[^a-z0-9_]+", "", key.lower())
        self[key.lower()] = value
        self[value] = key.lower()
        return value


protocol_translations = TranslationsDict(None)
[protocol_translations[protocol] for protocol in translate_protocols]


def serialize_packet_id(packet: PacketType) -> str:
    """Serialize packet identifiers into one reversible string.

    >>> serialize_packet_id({
    ...     'protocol': 'newkaku',
    ...     'id': '000001',
    ...     'switch': '01',
    ...     'command': 'on',
    ... })
    'newkaku_000001_01'
    >>> serialize_packet_id({
    ...     'protocol': 'ikea koppla',
    ...     'id': '000080',
    ...     'switch': '0',
    ...     'command': 'on',
    ... })
    'ikeakoppla_000080_0'
    >>> # unserializeable protocol name without explicit entry
    >>> # in translation table should be properly serialized
    >>> serialize_packet_id({
    ...     'protocol': 'alecto v4',
    ...     'id': '000080',
    ...     'switch': '0',
    ...     'command': 'on',
    ... })
    'alectov4_000080_0'
    """
    # translate protocol into something reversible
    protocol = protocol_translations[packet["protocol"]]

    if protocol == UNKNOWN:
        protocol = "rflink"

    return "_".join(
        filter(None, [protocol, packet.get("id", None), packet.get("switch", None)])
    )


def deserialize_packet_id(packet_id: str) -> Dict[str, str]:
    r"""Turn a packet id into individual packet components.

    >>> deserialize_packet_id('newkaku_000001_01') == {
    ...     'protocol': 'newkaku',
    ...     'id': '000001',
    ...     'switch': '01',
    ... }
    True
    >>> deserialize_packet_id('ikeakoppla_000080_0') == {
    ...     'protocol': 'ikea koppla',
    ...     'id': '000080',
    ...     'switch': '0',
    ... }
    True
    >>> deserialize_packet_id('dooya_v4_6d5f8e00_3f') == {
    ...     'protocol': 'dooya_v4',
    ...     'id': '6d5f8e00',
    ...     'switch': '3f',
    ... }
    True
    >>> deserialize_packet_id('mertik_gv60_038527_13') == {
    ...     'protocol': 'mertik_gv60',
    ...     'id': '038527',
    ...     'switch': '13',
    ... }
    True
    """
    if packet_id == "rflink":
        return {"protocol": UNKNOWN}

    # Protocol names can contain underscores themselves (like: dooya_v4), using rsplit to
    # prevent parsing issues with these kind of packets.
    protocol, *id_switch = packet_id.rsplit(PACKET_ID_SEP, 2)

    packet_identifiers = {
        # lookup the reverse translation of the protocol in the translation
        # table, fallback to protocol. If this is an unserializable protocol
        # name, it has not been serialized before and is not in the
        # translate_protocols table this will result in an invalid command.
        "protocol": protocol_translations.get(protocol, protocol),
    }
    if id_switch:
        packet_identifiers["id"] = id_switch[0]
    if len(id_switch) > 1:
        packet_identifiers["switch"] = id_switch[1]

    return packet_identifiers


def packet_events(packet: PacketType) -> Generator[PacketType, None, None]:
    """Return list of all events in the packet.

    >>> x = list(packet_events({
    ...     'protocol': 'alecto v1',
    ...     'id': 'ec02',
    ...     'temperature': 1.0,
    ...     'temperature_unit': '°C',
    ...     'humidity': 10,
    ...     'humidity_unit': '%',
    ... }))
    >>> assert {
    ...     'id': 'alectov1_ec02_temp',
    ...     'sensor': 'temperature',
    ...     'value': 1.0,
    ...     'unit': '°C',
    ... } in x
    >>> assert {
    ...     'id': 'alectov1_ec02_hum',
    ...     'sensor': 'humidity',
    ...     'value': 10,
    ...     'unit': '%',
    ... } in x
    >>> y = list(packet_events({
    ...     'protocol': 'newkaku',
    ...     'id': '000001',
    ...     'switch': '01',
    ...     'command': 'on',
    ... }))
    >>> assert {'id': 'newkaku_000001_01', 'command': 'on'} in y
    """
    field_abbrev = {
        v: k
        for k, v in sorted(
            PACKET_FIELDS.items(), key=lambda x: (x[1], x[0]), reverse=True
        )
    }

    packet_id = serialize_packet_id(packet)
    events = {f: v for f, v in packet.items() if f in field_abbrev}
    if "command" in events or "version" in events:
        # switch events only have one event in each packet
        yield dict(id=packet_id, **events)
    else:
        if packet_id == "debug":
            yield {
                "id": "raw",
                "value": packet.get("pulses(usec)"),
                "tm": packet.get("tm"),
                "pulses": packet.get("pulses"),
            }
        else:
            # sensors can have multiple
            for sensor, value in events.items():
                unit = packet.get(sensor + "_unit", None)
                yield {
                    "id": packet_id + PACKET_ID_SEP + field_abbrev[sensor],
                    "sensor": sensor,
                    "value": value,
                    "unit": unit,
                }

            if packet_id != "rflink":
                yield {
                    "id": packet_id + PACKET_ID_SEP + "update_time",
                    "sensor": "update_time",
                    "value": round(time.time()),
                    "unit": "s",
                }



"""Asyncio protocol implementation of RFlink."""

# ./.homeassistant/deps/lib/python/site-packages/rflink/protocol.py
# /Library/Frameworks/Python.framework/Versions/3.6//lib/python3.6/site-packages/rflink/protocol.py

import asyncio
import concurrent
import logging
from datetime import timedelta
from fnmatch import fnmatchcase
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

import socket

from serial_asyncio import create_serial_connection

from .parser import (
    PacketType,
    decode_packet,
    deserialize_packet_id,
    encode_packet,
    packet_events,
    valid_packet,
)

if TYPE_CHECKING:
    from typing import Coroutine  # not available in 3.4


log = logging.getLogger(__name__)
rflink_log = None

TIMEOUT = timedelta(seconds=5)
DEFAULT_TCP_KEEPALIVE_INTERVAL = 20
DEFAULT_TCP_KEEPALIVE_COUNT = 3


class ProtocolBase(asyncio.Protocol):
    """Manage low level rflink protocol."""

    transport = None  # type: asyncio.BaseTransport
    keepalive = None  # type: Optional[int]

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        disconnect_callback: Optional[Callable[[Optional[Exception]], None]] = None,
        keepalive: Optional[int] = None,
        **kwargs: Any
    ) -> None:
        """Initialize class."""
        if loop:
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()
        self.packet = ""
        self.buffer = ""
        self.packet_callback = None  # type: Optional[Callable[[PacketType], None]]
        self.disconnect_callback = disconnect_callback
        self.keepalive = keepalive

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Just logging for now."""
        self.transport = transport
        log.debug("connected")
        sock = transport.get_extra_info("socket")
        if self.keepalive is not None and socket is not None:
            log.debug(
                "applying TCP KEEPALIVE settings: IDLE={}/INTVL={}/CNT={}".format(
                    self.keepalive,
                    DEFAULT_TCP_KEEPALIVE_INTERVAL,
                    DEFAULT_TCP_KEEPALIVE_COUNT,
                )
            )
            if hasattr(socket, "SO_KEEPALIVE"):
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            if hasattr(socket, "TCP_KEEPIDLE"):
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, self.keepalive)
            if hasattr(socket, "TCP_KEEPINTVL"):
                sock.setsockopt(
                    socket.IPPROTO_TCP,
                    socket.TCP_KEEPINTVL,
                    DEFAULT_TCP_KEEPALIVE_INTERVAL,
                )
            if hasattr(socket, "TCP_KEEPCNT"):
                sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPCNT, DEFAULT_TCP_KEEPALIVE_COUNT
                )

    def data_received(self, data: bytes) -> None:
        """Add incoming data to buffer."""
        try:
            decoded_data = data.decode()
        except UnicodeDecodeError:
            invalid_data = data.decode(errors="replace")
            log.warning("Error during decode of data, invalid data: %s", invalid_data)
        else:
            log.debug("received data: %s", decoded_data.strip())
            self.buffer += decoded_data
            self.handle_lines()

    def handle_lines(self) -> None:
        """Assemble incoming data into per-line packets."""
        while "\r\n" in self.buffer:
            line, self.buffer = self.buffer.split("\r\n", 1)
            if valid_packet(line):
                self.handle_raw_packet(line)
            else:
                log.warning("dropping invalid data: %s", line)

    def handle_raw_packet(self, raw_packet: str) -> None:
        """Handle one raw incoming packet."""
        raise NotImplementedError()

    def send_raw_packet(self, packet: str) -> None:
        """Encode and put packet string onto write buffer."""
        data = packet + "\r\n"
        log.debug("writing data: %s", repr(data))
        # type ignore: transport from create_connection is documented to be
        # implementation specific bidirectional, even though typed as
        # BaseTransport
        self.transport.write(data.encode())  # type: ignore

    def log_all(self, file: Optional[str]) -> None:
        """Log all data received from RFLink to file."""
        global rflink_log
        if file is None:
            rflink_log = None
        else:
            log.debug("logging to: %s", file)
            rflink_log = open(file, "a")

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Log when connection is closed, if needed call callback."""
        if exc:
            log.exception("disconnected due to exception")
        else:
            log.info("disconnected because of close/abort.")
        if self.disconnect_callback:
            self.disconnect_callback(exc)


class PacketHandling(ProtocolBase):
    """Handle translating rflink packets to/from python primitives."""

    def __init__(
        self,
        *args: Any,
        packet_callback: Optional[Callable[[PacketType], None]] = None,
        **kwargs: Any
    ) -> None:
        """Add packethandling specific initialization.

        packet_callback: called with every complete/valid packet
        received.
        """
        super().__init__(*args, **kwargs)
        if packet_callback:
            self.packet_callback = packet_callback

    def handle_raw_packet(self, raw_packet: str) -> None:
        """Parse raw packet string into packet dict."""
        log.debug("got packet: %s", raw_packet)
        if rflink_log:
            print(raw_packet, file=rflink_log)
            rflink_log.flush()
        packet = None  # type: Optional[PacketType]
        try:
            packet = decode_packet(raw_packet)
        except BaseException:
            log.exception("failed to parse packet data: %s", raw_packet)

        log.debug("decoded packet: %s", packet)

        if packet:
            if "ok" in packet:
                # handle response packets internally
                log.debug("command response: %s", packet)
                self.handle_response_packet(packet)
            else:
                self.handle_packet(packet)
        else:
            log.warning("no valid packet")

    def handle_packet(self, packet: PacketType) -> None:
        """Process incoming packet dict and optionally call callback."""
        if self.packet_callback:
            # forward to callback
            self.packet_callback(packet)
        else:
            print("packet", packet)

    def handle_response_packet(self, packet: PacketType) -> None:
        """Handle response packet."""
        raise NotImplementedError()

    def send_packet(self, fields: PacketType) -> None:
        """Concat fields and send packet to gateway."""
        self.send_raw_packet(encode_packet(fields))

    def send_command(self, device_id: str, action: str) -> None:
        """Send device command to rflink gateway."""
        command = deserialize_packet_id(device_id)
        command["command"] = action
        log.debug("sending command: %s", command)
        self.send_packet(command)


class CommandSerialization(PacketHandling):
    """Logic for ensuring asynchronous commands are sent in order."""

    def __init__(
        self,
        *args: Any,
        packet_callback: Optional[Callable[[PacketType], None]] = None,
        **kwargs: Any
    ) -> None:
        """Add packethandling specific initialization."""
        super().__init__(*args, **kwargs)
        if packet_callback:
            self.packet_callback = packet_callback
        self._command_ack = asyncio.Event(loop=self.loop)
        self._ready_to_send = asyncio.Lock(loop=self.loop)

    def handle_response_packet(self, packet: PacketType) -> None:
        """Handle response packet."""
        self._last_ack = packet
        self._command_ack.set()

    @asyncio.coroutine
    def send_command_ack(
        self, device_id: str, action: str
    ) -> Generator[Any, None, Optional[bool]]:
        """Send command, wait for gateway to repond with acknowledgment."""
        # serialize commands
        yield from self._ready_to_send.acquire()
        acknowledgement = None
        try:
            self._command_ack.clear()
            self.send_command(device_id, action)

            log.debug("waiting for acknowledgement")
            try:
                yield from asyncio.wait_for(
                    self._command_ack.wait(), TIMEOUT.seconds, loop=self.loop
                )
                log.debug("packet acknowledged")
            except concurrent.futures._base.TimeoutError:
                acknowledgement = False
                log.warning("acknowledge timeout")
            else:
                acknowledgement = cast(bool, self._last_ack.get("ok", False))
        finally:
            # allow next command
            self._ready_to_send.release()

        return acknowledgement


class EventHandling(PacketHandling):
    """Breaks up packets into individual events with ids'.

    Most packets represent a single event (light on, measured
    temperature), but some contain multiple events (temperature and
    humidity). This class adds logic to convert packets into individual
    events each with their own id based on packet details (protocol,
    switch, etc).
    """

    def __init__(
        self,
        *args: Any,
        event_callback: Optional[Callable[[PacketType], None]] = None,
        ignore: Optional[Sequence[str]] = None,
        **kwargs: Any
    ) -> None:
        """Add eventhandling specific initialization."""
        super().__init__(*args, **kwargs)
        self.event_callback = event_callback
        # suppress printing of packets
        if not kwargs.get("packet_callback"):
            self.packet_callback = lambda x: None
        if ignore:
            log.debug("ignoring: %s", ignore)
            self.ignore = ignore
        else:
            self.ignore = []

    def _handle_packet(self, packet: PacketType) -> None:
        """Event specific packet handling logic.

        Break packet into events and fires configured event callback or
        nicely prints events for console.
        """
        events = packet_events(packet)

        for event in events:
            if self.ignore_event(event["id"]):
                log.debug("ignoring event with id: %s", event)
                continue
            log.debug("got event: %s", event)
            if self.event_callback:
                self.event_callback(event)
            else:
                self.handle_event(event)

    def handle_event(self, event: PacketType) -> None:
        """Handle of incoming event (print)."""
        string = "{id:<32} "
        if "command" in event:
            string += "{command}"
        elif "version" in event:
            if "hardware" in event:
                string += "{hardware} {firmware} "
            string += "V{version} R{revision}"
        else:
            string += "{value}"
            if event.get("unit"):
                string += " {unit}"

        print(string.format(**event))

    def handle_packet(self, packet: PacketType) -> None:
        """Apply event specific handling and pass on to packet handling."""
        self._handle_packet(packet)
        super().handle_packet(packet)

    def ignore_event(self, event_id: str) -> bool:
        """Verify event id against list of events to ignore.

        >>> e = EventHandling(ignore=[
        ...   'test1_00',
        ...   'test2_*',
        ... ])
        >>> e.ignore_event('test1_00')
        True
        >>> e.ignore_event('test2_00')
        True
        >>> e.ignore_event('test3_00')
        False
        """
        for ignore in self.ignore:
            if fnmatchcase(event_id, ignore):
                return True
        return False


class RflinkProtocol(CommandSerialization, EventHandling):
    """Combine preferred abstractions that form complete Rflink interface."""


class InverterProtocol(RflinkProtocol):
    """Invert switch commands received and send them out."""

    def handle_event(self, event: PacketType) -> None:
        """Handle incoming packet from rflink gateway."""
        if event.get("command"):
            if event["command"] == "on":
                cmd = "off"
            else:
                cmd = "on"

            task = self.send_command_ack(event["id"], cmd)
            self.loop.create_task(task)


class RepeaterProtocol(RflinkProtocol):
    """Repeat switch commands received."""

    def handle_event(self, packet: PacketType) -> None:
        """Handle incoming packet from rflink gateway."""
        if packet.get("command"):
            task = self.send_command_ack(packet["id"], packet["command"])
            self.loop.create_task(task)


@overload
def create_rflink_connection(
    port: int,
    host: str,
    baud: int = 57600,
    keepalive: Optional[int] = None,
    protocol: Type[ProtocolBase] = RflinkProtocol,
    packet_callback: Optional[Callable[[PacketType], None]] = None,
    event_callback: Optional[Callable[[PacketType], None]] = None,
    disconnect_callback: Optional[Callable[[Optional[Exception]], None]] = None,
    ignore: Optional[Sequence[str]] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> "Coroutine[Any, Any, Tuple[asyncio.BaseTransport, ProtocolBase]]":
    """Create Rflink manager class, returns transport coroutine."""
    ...


@overload
def create_rflink_connection(
    port: str,
    host: None = None,
    baud: int = 57600,
    keepalive: None = None,
    protocol: Type[ProtocolBase] = RflinkProtocol,
    packet_callback: Optional[Callable[[PacketType], None]] = None,
    event_callback: Optional[Callable[[PacketType], None]] = None,
    disconnect_callback: Optional[Callable[[Optional[Exception]], None]] = None,
    ignore: Optional[Sequence[str]] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> "Coroutine[Any, Any, Tuple[asyncio.BaseTransport, ProtocolBase]]":
    """Create Rflink manager class, returns transport coroutine."""
    ...


def create_rflink_connection(
    port: Union[None, str, int] = None,
    host: Optional[str] = None,
    baud: int = 57600,
    keepalive: Optional[int] = None,
    protocol: Type[ProtocolBase] = RflinkProtocol,
    packet_callback: Optional[Callable[[PacketType], None]] = None,
    event_callback: Optional[Callable[[PacketType], None]] = None,
    disconnect_callback: Optional[Callable[[Optional[Exception]], None]] = None,
    ignore: Optional[Sequence[str]] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> "Coroutine[Any, Any, Tuple[asyncio.BaseTransport, ProtocolBase]]":
    """Create Rflink manager class, returns transport coroutine."""
    if loop is None:
        loop = asyncio.get_event_loop()
    # use default protocol if not specified
    protocol_factory = partial(
        protocol,
        loop=loop,
        packet_callback=packet_callback,
        event_callback=event_callback,
        disconnect_callback=disconnect_callback,
        ignore=ignore if ignore else [],
        keepalive=keepalive,
    )

    # setup serial connection if no transport specified
    if host:
        conn = loop.create_connection(protocol_factory, host, cast(int, port))
    else:
        conn = create_serial_connection(loop, protocol_factory, port, baud)

    return conn  # type: ignore



"""Command line interface for rflink library.

Usage:
  rflink [-v | -vv] [options]
  rflink [-v | -vv] [options] [--repeat <repeat>] (on|off|allon|alloff|up|down|stop|pair) <id>
  rflink (-h | --help)
  rflink --version

Options:
  -p --port=<port>         Serial port to connect to [default: /dev/ttyACM0],
                             or TCP port in TCP mode.
  --baud=<baud>            Serial baud rate [default: 57600].
  --host=<host>            TCP mode, connect to host instead of serial port.
  --repeat=<repeat>        How often to repeat a command [default: 1].
  --keepalive=<NUMBER>     Enable TCP Keepalive IDLE timer in seconds.
  -m=<handling>            How to handle incoming packets [default: event].
  --ignore=<ignore>        List of device ids to ignore, wildcards supported.
  -h --help                Show this screen.
  -v                       Increase verbosity
  --version                Show version.
"""

import asyncio
import logging
import sys
from typing import Dict, Optional, Sequence, Type  # noqa: unused-import

import pkg_resources
from docopt import docopt

from .protocol import (  # noqa: unused-import
    CommandSerialization,
    EventHandling,
    InverterProtocol,
    PacketHandling,
    ProtocolBase,
    RepeaterProtocol,
    RflinkProtocol,
    create_rflink_connection,
)

PROTOCOLS = {
    "command": RflinkProtocol,
    "event": EventHandling,
    "print": PacketHandling,
    "invert": InverterProtocol,
    "repeat": RepeaterProtocol,
}  # type: Dict[str, Type[ProtocolBase]]

ALL_COMMANDS = ["on", "off", "allon", "alloff", "up", "down", "stop", "pair"]


def main(
    argv: Sequence[str] = sys.argv[1:], loop: Optional[asyncio.AbstractEventLoop] = None
) -> None:
    """Parse argument and setup main program loop."""
    args = docopt(
        __doc__, argv=argv, version=pkg_resources.require("rflink")[0].version
    )

    level = logging.ERROR
    if args["-v"]:
        level = logging.INFO
    if args["-v"] == 2:
        level = logging.DEBUG
    logging.basicConfig(level=level)

    if not loop:
        loop = asyncio.get_event_loop()

    if args["--ignore"]:
        ignore = args["--ignore"].split(",")
    else:
        ignore = []

    command = next((c for c in ALL_COMMANDS if args[c] is True), None)

    if command:
        protocol_type = PROTOCOLS["command"]
    else:
        protocol_type = PROTOCOLS[args["-m"]]

    keepalive = None
    if type(args["--keepalive"]) is str:
        keepalive = int(args["--keepalive"])

    conn = create_rflink_connection(
        protocol=protocol_type,
        host=args["--host"],
        port=args["--port"],
        baud=args["--baud"],
        loop=loop,
        ignore=ignore,
        keepalive=keepalive,
    )

    transport, protocol = loop.run_until_complete(conn)

    try:
        if command:
            assert isinstance(protocol, CommandSerialization)
            for _ in range(int(args["--repeat"])):
                loop.run_until_complete(
                    protocol.send_command_ack(args["<id>"], command)
                )
        else:
            loop.run_forever()
    except KeyboardInterrupt:
        # cleanup connection
        transport.close()
        loop.run_forever()
    finally:
        loop.close()


if __name__ == "__main__":
    # execute only if run as a script
    main()
